--[[
	This mod is largely based on the RANDTECH mod and provides an alternative approach on how the techtree is randomized. 
	For each cost-category, e.g. all tech that cost 900 RP, the tech are shuffled randomly among each other. 
	While this does affect balance (e.g. certain 900 RP areas are reachable faster than others) it does not change the 
	individual cost of any tech (this is the main difference to RANDTECH). 
	
	The implementation is rather crude but performance is not an issue because the generation is done only once at the start of 
	the game and takes less than a second. 
	
	The mod seems to work with all starting eras and uncreative races but there could still be some edge cases which cause issues, 
	such as the auto-removal of shield-tech if no shield-class is available for an uncreative race. 
	
	Last update of this mod: 20th February 2021; Tested with version 1.50.13 improved
	
	Questions etc. to: romanhable@web.de (Epirasque)
]]

advf = {
    advf_biology = 1,
    advf_power = 1,
    advf_physics = 1,
    advf_construction = 1,
    advf_fields = 1,
    advf_chemistry = 1,
    advf_computers = 1,
    advf_sociology = 1
}

function set_id(t, tag)
    for i,v in pairs(t) do
        v[tag] = i
    end
end

function make_branches(ff, keep_gray_boxes)
    local branches = {}
    local f2br = {}
    local starting_fields = rget_starting_fields()

    local map = { engineering = "advanced_engineering" }
    if keep_gray_boxes then
        map.chemistry = "advanced_metallurgy"
        map.nuclear_fission = "advanced_fusion"
        map.physics = "fusion_physics"
        map.electronics = "optronics"
    end
    for i,s in pairs(starting_fields) do
        if map[s] ~= nil then
            starting_fields[i] = map[s]
        end
    end

    for i,f in pairs(starting_fields) do
        local n = rawlen(branches) + 1
        local b = {
            branch_last_field = "advf_" .. i,
            branchnum = n,
            field_id16 = f,
        }
        table.insert(branches, b)
        ff[f].is_branch_root = 1
    end

    for i,b in ipairs(branches) do
        local fname = b.field_id16
        local marks = {}
        local n = 1
        while advf[fname] == nil and ff[fname] ~= nil do
            local f = ff[fname]
            if marks[fname] ~= nil then
                error(tostring("detected tech loop ", marks))
            end
            marks[fname] = n
            n = n + 1
            f.branch = i
            --f.techs = {}
            fname = f.next_field
        end
    end

    return branches
end

function ordered_fields_by_cost(ff)
    local fbycost = {}
    for i,f in pairs(ff) do
        if f.branch ~= nil and f.is_branch_root == nil then
            table.insert(fbycost, f)
        end
    end
    table.sort(fbycost, function(a,b) return a.cost < b.cost end)
    return fbycost
end

function ordered_techs_by_cost(tt, ff)
    local tbycost = {}
    for i,t in pairs(tt) do
        if t.field_id16 ~= "advanced_governments" then
            local f = ff[t.field_id16]
            if f ~= nil and f.branch ~= nil then
                t.cost = f.cost
                table.insert(tbycost, t)
            end
        end
    end
    table.sort(tbycost, function(a,b) return a.cost < b.cost end)
    return tbycost
end

function getTechsOfSameCost(allTechs, referenceTech)
	if referenceTech.cost == nil then
		return {}
	end
	local candidateTechs = {}
	for candidateTechID, tech in pairs(allTechs) do
		if tech.cost == referenceTech.cost then
			table.insert(candidateTechs, candidateTechID)
		end
	end
	return candidateTechs
end


function switchAllTechs(tt, ff)
	for techToSwitchID,techToSwitch in pairs(tt) do

		local techsOfSameCost = getTechsOfSameCost(tt, techToSwitch)
		
		if techToSwitch ~= nil and techToSwitch.cost ~= nil and random(10) <= 7 then
			if #techsOfSameCost == 0 then
				msgbox("no same cost for " .. tostring(techToSwitch))
			end
			
			local switchTargetID = techsOfSameCost[ random(#techsOfSameCost) ]
			
			local tmpField = tt[switchTargetID]
			tt[switchTargetID] = tt[techToSwitchID]
			tt[techToSwitchID] = tmpField
		    
		end
	end
end

function go()
    set_seed()
    local ff = rget_fields() set_id(ff, "name")
    local tt = rget_techs()  set_id(tt, "name")
	local bb = make_branches(ff, 1)
    local fbc = ordered_fields_by_cost(ff)
    local tbc = ordered_techs_by_cost(tt, ff)
	
	switchAllTechs(tt, ff)
	switchAllTechs(tt, ff)
	switchAllTechs(tt, ff)
	
    rset_fields(ff)
    rset_techs(tt)
end

go()
msgbox("Tech tree has been randomized without changing RP costs")
